# Licensed under the terms of http://www.apache.org/licenses/LICENSE-2.0
# Author (©): Alvaro del Castillo
import logging
import math

from mcpi.vec3 import Vec3
import mcpi.block

from mcthings.utils import size_region, find_min_max_cuboid_vertex


class BlockMemory:

    def __init__(self, block_id, block_data, pos):
        self.id = block_id
        self.data = block_data
        self.pos = pos


class BlocksMemory:
    """
    Blocks memory for a Thing
    """

    def __init__(self):

        self.blocks = []
        self._blocks_pos = {}

    def add(self, block_memory):
        """
        Add a new block to the memory.

        :param block_memory: memory for a block
        :return:
        """

        self.blocks.append(block_memory)

    def find_init_end_pos(self):
        """ Find the init and end cuboid positions from all the blocks in the memory """

        first_pos = self.blocks[0].pos

        init_pos = Vec3(first_pos.x, first_pos.y, first_pos.z)
        end_pos = Vec3(first_pos.x, first_pos.y, first_pos.z)

        for block in self.blocks:
            pos = block.pos
            if pos.x < init_pos.x:
                init_pos = Vec3(pos.x, init_pos.y, init_pos.z)
            if pos.y < init_pos.y:
                init_pos = Vec3(init_pos.x, pos.y, init_pos.z)
            if pos.z < init_pos.z:
                init_pos = Vec3(init_pos.x, init_pos.y, pos.z)

            if pos.x > end_pos.x:
                end_pos = Vec3(pos.x, end_pos.y, end_pos.z)
            if pos.y > end_pos.y:
                end_pos = Vec3(end_pos.x, pos.y, end_pos.z)
            if pos.z > end_pos.z:
                end_pos = Vec3(end_pos.x, end_pos.y, pos.z)

        return init_pos, end_pos

    def is_cuboid(self):
        """ Check if the memory is a filled cuboid """

        cuboid = False

        # Check that the number of blocks needed for the filled cuboid is the same that the blocks
        init_pos, vertex_max = self.find_init_end_pos()
        size = size_region(init_pos, vertex_max)

        if size.x * size.y * size.z == len(self.blocks):
            cuboid = True

        return cuboid

    def memory_equal(self):
        """ Check if all the blocks in the memory are equal """
        equal = True

        if self.blocks:
            last_block = self.blocks[0]

            for block in self.blocks:
                if block.id != last_block.id or block.data != last_block.data:
                    equal = False
                    break
                last_block = block
        else:
            equal = False

        return equal

    def flip_x(self, position):
        """
        Flip based on x-axis the blocks in memory using position as base position from which to rotate
        :param position: base position from which to rotate
        :return:
        """

        for block in self.blocks:
            # Find the x position and flip it
            width = abs(block.pos.x - position.x)
            # TODO: the flip could be done in two directions (left or right)
            # This one the the flip to the right
            x_flipped = position.x - width
            block.pos.x = x_flipped

    def fill(self, fill_block):
        """
        Fill all blocks in memory with fill_block

        :param fill_block: block to be used to fill all memory
        :return:
        """

        for block in self.blocks:
            block.id = fill_block.id
            block.data = fill_block.data

    def rotate(self, degrees, position):
        """
        Rotate degrees the blocks in memory using position as base position from which to rotate
        :param degrees: degrees to rotate (90, 180, 270)
        :param position: base position from which to rotate
        :return:
        """
        valid_degrees = [90, 180, 270]

        if degrees not in [90, 180, 270]:
            raise RuntimeError("Invalid degrees: %s (valid: %s) " % (degrees, valid_degrees))

        cos_degrees = math.cos(math.radians(degrees))
        sin_degrees = math.sin(math.radians(degrees))

        def rotate_x(pos_x, pos_z):
            return pos_x * cos_degrees - pos_z * sin_degrees

        def rotate_z(pos_x, pos_z):
            return pos_z * cos_degrees + pos_x * sin_degrees

        # Base position for the rotation
        init_pos = position
        rotated_blocks = []

        # Rotate all blocks with respect the initial position and add them
        for block in self.blocks:
            b = block.id
            d = block.data

            x = block.pos.x - init_pos.x
            z = block.pos.z - init_pos.z
            rotated_x = round(init_pos.x + rotate_x(x, z))
            rotated_z = round(init_pos.z + rotate_z(x, z))
            rotated_blocks.append(BlockMemory(b, d, Vec3(rotated_x, block.pos.y, rotated_z)))

        # Replace all blocks in memory with the rotated ones
        self.blocks = []
        for rotated_block in rotated_blocks:
            self.set_block(rotated_block.pos, rotated_block.id, rotated_block.data)

    def set_block(self, pos, block_id, block_data=None):
        self.add(BlockMemory(block_id, block_data, pos))

    def set_blocks(self, vertex, vertex_opposite, block_id):
        """ Add a cuboid with the same block for all blocks and without specific data """

        block_data = None

        width = abs(vertex_opposite.x - vertex.x) + 1
        height = abs(vertex_opposite.y - vertex.y) + 1
        length = abs(vertex_opposite.z - vertex.z) + 1

        vertex_min, vertex_max = find_min_max_cuboid_vertex(vertex, vertex_opposite)

        for y in range(0, height):
            for z in range(0, length):
                for x in range(0, width):
                    block_pos = Vec3(vertex_min.x + x, vertex_min.y + y, vertex_min.z + z)
                    self.set_block(block_pos, block_id, block_data)

    def _create_blocks_pos(self):
        logging.info("Creating the memory cache with positions")
        for block in self.blocks:
            self._blocks_pos[str(block.pos)] = block
        logging.info("Done memory cache with positions")

    def find_block_at_pos(self, pos):
        """
        Find a block in memory give its position
        TODO: Improve performance

        :param pos: position for the block
        :return: the block found or None
        """

        if not self._blocks_pos:
            self._create_blocks_pos()

        block_found = None
        if str(pos) in self._blocks_pos:
            block_found = self._blocks_pos[str(pos)]

        return block_found

    def to_nbt(self, init_pos, end_pos):
        """
        Convert the blocks of memory to NBT format for exporting as Schematic
        The NBT must be a complete cuboid with air in the positions where
        there are no data in blocks memory.


        :return: bytearrays for blocks ids and block data
        """

        size = size_region(init_pos, end_pos)

        blocks_bytes = bytearray()
        data_bytes = bytearray()

        # Use the same loop than reading Schematic format: x -> z -> y
        for y in range(0, size.y):
            for z in range(0, size.z):
                for x in range(0, size.x):
                    block_data = 0
                    block_id = mcpi.block.AIR.id
                    block_pos = Vec3(init_pos.x + x, init_pos.y + y, init_pos.z + z)
                    # Find if there is a block at block_pos
                    mem_block = self.find_block_at_pos(block_pos)
                    if mem_block:
                        block_id = mem_block.id
                        block_data = mem_block.data
                    blocks_bytes.append(block_id)
                    data_bytes.append(block_data)

        return blocks_bytes, data_bytes

    def build_schematic(self):
        init_pos, end_pos = self.find_init_end_pos()

        return self.to_nbt(init_pos, end_pos, self)

    def to_schematic(self, file_path):
        """
        Convert the blocks memory to a Schematic Object

        :file_path: file in which to export the memory in Schematic format
        :return: the Schematic object
        """

        self.build_schematic().write_file(file_path)




# Licensed under the terms of http://www.apache.org/licenses/LICENSE-2.0
# Author/s (©): Alvaro del Castillo

from math import sqrt

import chunk
import logging

import mcpi.block
from mcpi.vec3 import Vec3

from mcthings.thing import Thing


class Voxel:
    def __init__(self, bytes):
        self.x = bytes[0]
        self.y = bytes[1]
        self.z = bytes[2]
        self.color_index = bytes[3] - 1


class Color:
    """ RGBA format palette """
    _color2minecraft = {}  # Cache to convert a color to minecraft color

    def __init__(self, hex_str):
        self.hex_str = hex_str

    def rgb(self):
        red = int(self.hex_str[0:2], 16)
        green = int(self.hex_str[2:4], 16)
        blue = int(self.hex_str[4:6], 16)

        return red, green, blue

    def compare(self, color):
        # https://www.compuphase.com/cmetric.htm
        r1, g1, b1 = self.rgb()
        r2, g2, b2 = color.rgb()

        read_mean = (r1 + r2) / 2
        r = r1 - r2
        g = g1 - g2
        b = b1 - b2
        return sqrt((round((512 + read_mean) * r * r) >> 8) + 4.0 * g * g + (round((767 - read_mean) * b * b) >> 8))

    def minecraft(self):
        # https://gaming.stackexchange.com/questions/47212/what-are-the-color-values-for-dyed-wool
        mc_colors = [
            ("White", "e4e4e4"),
            ("Orange", "ea7e35"),
            ("Magenta", "be49c9"),
            ("Light Blue", "6387d2"),
            ("Yellow", "c2b51c"),
            ("Lime", "39ba2e"),
            ("Pink", "d98199"),
            ("Grey", "414141"),
            ("Light grey", "a0a7a7"),
            ("Cyan", "267191"),
            ("Purple", "7e34bf"),
            ("Blue", "253193"),
            ("Brown", "56331c"),
            ("Green", "364b18"),
            ("Red", "9e2b27"),
            ("Black", "181414")
        ]

        mc_color_number = {}
        mc_colors_hex = {}
        for i in range(0, len(mc_colors)):
            mc_color_number[mc_colors[i][1]] = i
            mc_colors_hex[mc_colors[i][1]] = mc_colors[i][0]

        # Find the closest Minecraft color
        rgb = self.hex_str[0:6]
        dist = float('inf')
        if rgb in mc_colors_hex:
            # Direct mapping
            color = rgb
        elif rgb in Color._color2minecraft:
            # Color already mapped
            color = Color._color2minecraft[rgb]
        else:
            for mc_color in mc_colors_hex:
                cdist = self.compare(Color(mc_color))
                if cdist < dist:
                    dist = cdist
                    color = mc_color
            self._color2minecraft[rgb] = color

        return mc_color_number[color]


class VoxDefaultPalette:
    # Removed first "0x00000000" (it does not appear in MV)
    # Reverse order: ABGR
    palette = [
        "0xffffffff", "0xffccffff", "0xff99ffff", "0xff66ffff", "0xff33ffff", "0xff00ffff", "0xffffccff",
        "0xffccccff",
        "0xff99ccff", "0xff66ccff", "0xff33ccff", "0xff00ccff", "0xffff99ff", "0xffcc99ff", "0xff9999ff",
        "0xff6699ff", "0xff3399ff", "0xff0099ff", "0xffff66ff", "0xffcc66ff", "0xff9966ff", "0xff6666ff", "0xff3366ff",
        "0xff0066ff",
        "0xffff33ff", "0xffcc33ff", "0xff9933ff", "0xff6633ff", "0xff3333ff", "0xff0033ff", "0xffff00ff",
        "0xffcc00ff", "0xff9900ff", "0xff6600ff", "0xff3300ff", "0xff0000ff", "0xffffffcc", "0xffccffcc", "0xff99ffcc",
        "0xff66ffcc",
        "0xff33ffcc", "0xff00ffcc", "0xffffcccc", "0xffcccccc", "0xff99cccc", "0xff66cccc", "0xff33cccc",
        "0xff00cccc", "0xffff99cc", "0xffcc99cc", "0xff9999cc", "0xff6699cc", "0xff3399cc", "0xff0099cc", "0xffff66cc",
        "0xffcc66cc",
        "0xff9966cc", "0xff6666cc", "0xff3366cc", "0xff0066cc", "0xffff33cc", "0xffcc33cc", "0xff9933cc",
        "0xff6633cc", "0xff3333cc", "0xff0033cc", "0xffff00cc", "0xffcc00cc", "0xff9900cc", "0xff6600cc", "0xff3300cc",
        "0xff0000cc",
        "0xffffff99", "0xffccff99", "0xff99ff99", "0xff66ff99", "0xff33ff99", "0xff00ff99", "0xffffcc99",
        "0xffcccc99", "0xff99cc99", "0xff66cc99", "0xff33cc99", "0xff00cc99", "0xffff9999", "0xffcc9999", "0xff999999",
        "0xff669999",
        "0xff339999", "0xff009999", "0xffff6699", "0xffcc6699", "0xff996699", "0xff666699", "0xff336699",
        "0xff006699", "0xffff3399", "0xffcc3399", "0xff993399", "0xff663399", "0xff333399", "0xff003399", "0xffff0099",
        "0xffcc0099",
        "0xff990099", "0xff660099", "0xff330099", "0xff000099", "0xffffff66", "0xffccff66", "0xff99ff66",
        "0xff66ff66", "0xff33ff66", "0xff00ff66", "0xffffcc66", "0xffcccc66", "0xff99cc66", "0xff66cc66", "0xff33cc66",
        "0xff00cc66",
        "0xffff9966", "0xffcc9966", "0xff999966", "0xff669966", "0xff339966", "0xff009966", "0xffff6666",
        "0xffcc6666", "0xff996666", "0xff666666", "0xff336666", "0xff006666", "0xffff3366", "0xffcc3366", "0xff993366",
        "0xff663366",
        "0xff333366", "0xff003366", "0xffff0066", "0xffcc0066", "0xff990066", "0xff660066", "0xff330066",
        "0xff000066", "0xffffff33", "0xffccff33", "0xff99ff33", "0xff66ff33", "0xff33ff33", "0xff00ff33", "0xffffcc33",
        "0xffcccc33",
        "0xff99cc33", "0xff66cc33", "0xff33cc33", "0xff00cc33", "0xffff9933", "0xffcc9933", "0xff999933",
        "0xff669933", "0xff339933", "0xff009933", "0xffff6633", "0xffcc6633", "0xff996633", "0xff666633", "0xff336633",
        "0xff006633",
        "0xffff3333", "0xffcc3333", "0xff993333", "0xff663333", "0xff333333", "0xff003333", "0xffff0033",
        "0xffcc0033", "0xff990033", "0xff660033", "0xff330033", "0xff000033", "0xffffff00", "0xffccff00", "0xff99ff00",
        "0xff66ff00",
        "0xff33ff00", "0xff00ff00", "0xffffcc00", "0xffcccc00", "0xff99cc00", "0xff66cc00", "0xff33cc00",
        "0xff00cc00", "0xffff9900", "0xffcc9900", "0xff999900", "0xff669900", "0xff339900", "0xff009900", "0xffff6600",
        "0xffcc6600",
        "0xff996600", "0xff666600", "0xff336600", "0xff006600", "0xffff3300", "0xffcc3300", "0xff993300",
        "0xff663300", "0xff333300", "0xff003300", "0xffff0000", "0xffcc0000", "0xff990000", "0xff660000", "0xff330000",
        "0xff0000ee",
        "0xff0000dd", "0xff0000bb", "0xff0000aa", "0xff000088", "0xff000077", "0xff000055", "0xff000044",
        "0xff000022", "0xff000011", "0xff00ee00", "0xff00dd00", "0xff00bb00", "0xff00aa00", "0xff008800", "0xff007700",
        "0xff005500",
        "0xff004400", "0xff002200", "0xff001100", "0xffee0000", "0xffdd0000", "0xffbb0000", "0xffaa0000",
        "0xff880000", "0xff770000", "0xff550000", "0xff440000", "0xff220000", "0xff110000", "0xffeeeeee", "0xffdddddd",
        "0xffbbbbbb",
        "0xffaaaaaa", "0xff888888", "0xff777777", "0xff555555", "0xff444444", "0xff222222", "0xff111111"
    ]


class Vox(Thing):
    file_path = None
    """ file path for the MagicaVoxel vox file """

    def parse_vox_file(self):
        if not self.file_path:
            RuntimeError("Missing file_path param")

        self.voxels = []
        self.palette = []
        self.materials = []

        # Read the vox data in RIFF format
        # https://github.com/python/cpython/blob/3.8/Lib/chunk.py
        vox_file = open(self.file_path, "rb")
        vox_chunk = chunk.Chunk(vox_file, bigendian=False)
        chunk_name = vox_chunk.getname().decode("utf-8")
        if chunk_name != 'VOX ':
            raise RuntimeError('File %s is not a VOX file' % self.file_path)
        version = vox_chunk.chunksize
        if version != 150:
            raise RuntimeError('File %s has a not supported VOX version %i' % (self.file_path, version))
        # Let's read chunks
        """
        2. Chunk Structure
        -------------------------------------------------------------------------------
        # Bytes  | Type       | Value
        -------------------------------------------------------------------------------
        1x4      | char       | chunk id
        4        | int        | num bytes of chunk content (N)
        4        | int        | num bytes of children chunks (M)      
        N        |            | chunk content
        M        |            | children chunks
        -------------------------------------------------------------------------------
        """
        # MAIN Chunk
        main_chunk = chunk.Chunk(vox_file, bigendian=False)
        # Pass last 4 bytes for MAIN Chunk with children chunks
        vox_file.seek(vox_file.tell() + 4)

        # SIZE CHUNK
        """
        -------------------------------------------------------------------------------
        # Bytes  | Type       | Value
        -------------------------------------------------------------------------------
        4        | int        | size x
        4        | int        | size y
        4        | int        | size z : gravity direction
        -------------------------------------------------------------------------------
        """
        size_chunk = chunk.Chunk(vox_file, bigendian=False)
        vox_file.seek(vox_file.tell() + 4)  # number of children chunks
        x = size_chunk.read(4)
        y = size_chunk.read(4)
        z = size_chunk.read(4)

        # XYZI voxels
        """
        -------------------------------------------------------------------------------
        # Bytes  | Type       | Value
        -------------------------------------------------------------------------------
        4        | int        | numVoxels (N)
        4 x N    | int        | (x, y, z, colorIndex) : 1 byte for each component
        -------------------------------------------------------------------------------
        """
        xyzi_chunk = chunk.Chunk(vox_file, bigendian=False)
        vox_file.seek(vox_file.tell() + 4)  # number of children chunks
        n_voxels_bytes = xyzi_chunk.read(4)
        n_voxels = int.from_bytes(n_voxels_bytes, "little")
        for i in range(0, n_voxels):
            self.voxels.append(Voxel(xyzi_chunk.read(4)))
        # Transform or palette chunk or no more chunk if default palette
        try:
            transform_chunk = chunk.Chunk(vox_file, bigendian=False)
        except EOFError:
            transform_chunk = None
            logging.info("Legacy vox file with default palette")
        if transform_chunk and transform_chunk.chunkname.decode("utf-8") == "nTRN":
            vox_file.seek(vox_file.tell() + 4)  # number of children chunks
            transform_chunk.skip()

            group_chunk = chunk.Chunk(vox_file, bigendian=False)
            vox_file.seek(vox_file.tell() + 4)  # number of children chunks
            group_chunk.skip()

            transform_chunk = chunk.Chunk(vox_file, bigendian=False)
            vox_file.seek(vox_file.tell() + 4)  # number of children chunks
            # transform_chunk.skip()  # it is skipping 1 byte in the next chunk
            vox_file.read(transform_chunk.getsize())

            shape_chunk = chunk.Chunk(vox_file, bigendian=False)
            vox_file.seek(vox_file.tell() + 4)  # number of children chunks
            vox_file.read(shape_chunk.getsize())

            # Layers chunk
            NUM_LAYERS = 8  # By default
            for i in range(0, NUM_LAYERS):
                layer_chunk = chunk.Chunk(vox_file, bigendian=False)
                vox_file.seek(vox_file.tell() + 4)  # number of children chunks
                vox_file.read(layer_chunk.getsize())
            """
            7. Chunk id 'RGBA' : palette
            -------------------------------------------------------------------------------
            # Bytes  | Type       | Value
            -------------------------------------------------------------------------------
            4 x 256  | int        | (R, G, B, A) : 1 byte for each component
                                  | * <NOTICE>
                                  | * color [0-254] are mapped to palette index [1-255], e.g : 
                                  | 
                                  | for ( int i = 0; i <= 254; i++ ) {
                                  |     palette[i + 1] = ReadRGBA(); 
                                  | }
            -------------------------------------------------------------------------------
            """
            rgba_chunk = chunk.Chunk(vox_file, bigendian=False)
            vox_file.seek(vox_file.tell() + 4)  # number of children chunks
        else:
            rgba_chunk = transform_chunk
            vox_file.seek(vox_file.tell() + 4)  # notice
        if rgba_chunk:
            if rgba_chunk.getname().decode("utf-8") != 'RGBA':
                raise RuntimeError('VOX format not supported (multimodel?)')
            for i in range(0, round(rgba_chunk.getsize() / 4)):
                # RGBA
                color_bytes = rgba_chunk.read(1) + rgba_chunk.read(1) + rgba_chunk.read(1) + rgba_chunk.read(1)
                self.palette.append(Color(color_bytes.hex()))
        else:
            # Default palette
            for i in range(0, len(VoxDefaultPalette.palette)):
                color = VoxDefaultPalette.palette[i].replace('0x', '')
                # Convert ABGR to RGBA
                color = color[::-1]
                self.palette.append(Color(color))

        # Read the materials palette
        """
        (4) Material Chunk : "MATL"

        int32	: material id
        DICT	: material properties
                    (_type : str) _diffuse, _metal, _glass, _emit
                    (_weight : float) range 0 ~ 1
                    (_rough : float)
                    (_spec : float)
                    (_ior : float)
                    (_att : float)
                    (_flux : float)
                    (_plastic)
        """
        # One material per each color
        for i in range(0, len(self.palette)):
            try:
                materials_chunk = chunk.Chunk(vox_file, bigendian=False)
                if materials_chunk.getname().decode("utf-8") != 'MATL':
                    logging.info("Material data not found")
                    break
                else:
                    vox_file.seek(vox_file.tell() + 4)  # number of children chunks
                    material_id = int.from_bytes(materials_chunk.read(4), "little")
                    dict_entries_len = int.from_bytes(materials_chunk.read(4), "little")
                    # Read the _type key from dict
                    key_str_len = int.from_bytes(materials_chunk.read(4), "little")
                    key_str = materials_chunk.read(key_str_len).decode('utf-8')
                    value_str_len = int.from_bytes(materials_chunk.read(4), "little")
                    value_str = materials_chunk.read(value_str_len).decode('utf-8')
                    self.materials.append(value_str)
                    materials_chunk.skip()
                    if materials_chunk.tell() > materials_chunk.getsize():
                        vox_file.seek(vox_file.tell() - 1)  # Hack: not sure why skip goes 1 byte more
            except EOFError:
                logging.info("Material data not found")
                break

    @classmethod
    def find_minecraft_material(cls, material):
        mc_material = None
        if material == '_glass':
            mc_material = mcpi.block.GLASS
        elif material == '_metal':
            mc_material = mcpi.block.IRON_BLOCK
        return mc_material

    def create(self):

        self.parse_vox_file()

        for voxel in self.voxels:
            voxel_color = self.palette[voxel.color_index]
            minecraft_material = None
            if self.materials:
                minecraft_material = self.find_minecraft_material(self.materials[voxel.color_index])
            minecraft_color = voxel_color.minecraft()

            # y, z are the reverse in vox format
            pos = Vec3(self.position.x + voxel.x,
                       self.position.y + voxel.z,
                       self.position.z + voxel.y
                      )

            if self.block == self._block_empty:
                self.set_block(pos, self._block_empty)
            elif minecraft_material:
                self.set_block(pos, minecraft_material.id)
            else:
                self.set_block(pos, mcpi.block.WOOL.id, minecraft_color)

        init_pos, end_pos = self._blocks_memory.find_init_end_pos()
        self._end_position = end_pos



# Licensed under the terms of http://www.apache.org/licenses/LICENSE-2.0
# Author (©): Alvaro del Castillo

from mcpi.vec3 import Vec3

from .thing import Thing


class Block(Thing):

    def create(self):
        self.set_block(Vec3(self.position.x, self.position.y, self.position.z), self.block.id)
        self._end_position = self.position



# Licensed under the terms of http://www.apache.org/licenses/LICENSE-2.0
# Author/s (©): Alvaro del Castillo

import mcpi.block
from nbt import nbt

from mcpi.vec3 import Vec3

from mcthings.thing import Thing


class Schematic(Thing):
    _blocks_field = 'Blocks'
    _data_field = 'Data'
    file_path = None
    """ file path for the schematic file """
    rotate_degrees = 0
    """ rotate the schematic """
    change_blocks = {mcpi.block.AIR.id: mcpi.block.AIR.id}
    """ Change a block with other """

    def find_bounding_box(self):
        """ In a Schematic the bounding box is inside the file data """

        schematic = nbt.NBTFile(self.file_path, 'rb')

        size_x = schematic["Width"].value
        size_y = schematic["Height"].value
        size_z = schematic["Length"].value

        init_x = self.position.x
        init_y = self.position.y
        init_z = self.position.z

        self._end_position = Vec3(init_x + size_x - 1,
                                  init_y + size_y - 1,
                                  init_z + size_z - 1)

        return self.position, self.end_position

    def create(self):
        if not self.file_path:
            RuntimeError("Missing file_path param")

        schematic = nbt.NBTFile(self.file_path, 'rb')
        size_x = schematic["Width"].value
        size_y = schematic["Height"].value
        size_z = schematic["Length"].value

        init_pos = self.position

        for y in range(0, size_y):
            for z in range(0, size_z):
                for x in range(0, size_x):
                    i = x + size_x * z + (size_x * size_z) * y
                    block_id = schematic[self._blocks_field][i]
                    block_data = schematic[self._data_field][i] & 0b00001111  # lower 4 bits
                    block_pos = Vec3(init_pos.x + x, init_pos.y + y, init_pos.z + z)
                    if block_id in self.change_blocks:
                        block_id = self.change_blocks[block_id]
                    self.set_block(block_pos, block_id, block_data)

        init_pos, self._end_position = self.find_bounding_box()



# Licensed under the terms of http://www.apache.org/licenses/LICENSE-2.0
# Author (©): Alvaro del Castillo

import mcpi.block
from mcpi.vec3 import Vec3

from .thing import Thing
from .world import World


class River(Thing):
    length = 100
    width = 2
    depth = 1
    block = mcpi.block.WATER_FLOWING

    def create(self):
        init_x = self.position.x
        init_y = self.position.y - self.depth
        init_z = self.position.z

        end_x = init_x + self.width - 1
        end_y = self.position.y - 1
        end_z = init_z + self.length - 1

        # Find the type of land block destroyed with the river
        self._block_empty = \
            mcpi.block.Block(World.renderer.get_block(Vec3(init_x, self.position.y - 1, init_z)), 0)

        self.set_blocks(
            Vec3(init_x, init_y, init_z),
            Vec3(end_x, end_y, end_z),
            self.block.id)

        self._end_position = Vec3(end_x, self.position.y - self.depth, end_z)



# Licensed under the terms of http://www.apache.org/licenses/LICENSE-2.0
# Author (©): Alvaro del Castillo

import mcpi

from mcpi.vec3 import Vec3
import mcpi.block

from .thing import Thing


class Wool(Thing):
    # Wool colors
    COLORS = [
        "White",
        "Orange",
        "Magenta",
        "Light Blue",
        "Yellow",
        "Lime",
        "Pink",
        "Grey",
        "Light grey",
        "Cyan",
        "Purple",
        "Blue",
        "Brown",
        "Green",
        "Red",
        "Black"
    ]

    def create(self):
        """
        Show all wool colors
        :return:
        """

        for i in range(0, len(self.COLORS)):
            p = self.position
            self.set_block(Vec3(p.x + i, p.y, p.z), mcpi.block.WOOL.id, i)

        self._end_position = Vec3(p.x + len(self.COLORS) - 1, p.y, p.z)



# Licensed under the terms of http://www.apache.org/licenses/LICENSE-2.0
# Author (©): Alvaro del Castillo

# TODO: at some point this must be a real Singleton


class World:
    """
    A world is a container for all the scenes built using McThings. Its mapping
    is direct to Minecraft world concept.

    Before adding Things to the World, it must have a renderer
    """
    scenes = []
    """ Scenes included in the world """
    renderer = None
    """ Render used to render the scenes """

    @classmethod
    def set_renderer(cls, renderer):
        cls.renderer = renderer

        # TODO: Hack for Minecraft renderer to use McDrawing
        from mcthings.renderers.raspberry_pi import RaspberryPi
        if isinstance(renderer, RaspberryPi):
            World.drawing = renderer.drawing

    @classmethod
    def add_scene(cls, scene):
        """ Add a new scene to the world """
        cls.scenes.append(scene)

    @classmethod
    def first_scene(cls):
        """ Return the first scene used be default """
        return cls.scenes[0]

    @classmethod
    def build(cls):
        """ Build all the scenes inside the world """
        for scene in cls.scenes:
            scene.build()

    @classmethod
    def unbuild(cls):
        """ Unbuild all the scenes inside the world """
        for scene in cls.scene:
            scene.unbuild()



# Licensed under the terms of http://www.apache.org/licenses/LICENSE-2.0
# Author (©): Alvaro del Castillo

import mcpi.block
from mcpi.vec3 import Vec3

from ._version import __version__

from .blocks_memory import BlocksMemory
from .scene import Scene
from .utils import build_schematic_nbt
from .world import World


class Thing:
    """ base class for all objects in mcthings library """

    block = mcpi.block.BRICK_BLOCK
    """ block type used by the thing. Default to BRICK_BLOCK """
    _block_empty = mcpi.block.AIR
    """ block type used to remove blocks in this Thing """

    def __init__(self, position, parent=None, scene=None):
        """
        Create a thing
        :param position: build position
        :param parent: parent Thing in which this one is included
        :param scene: scene in which this Thing is included
        """

        self._blocks_memory = BlocksMemory()
        self._children = []
        self._decorators = []
        self._end_position = None
        self._parent = parent
        self._position = None
        self._scene = scene

        if position:
            if not (isinstance(position.x, int) and
                    isinstance(position.y, int) and
                    isinstance(position.z, int)):
                raise RuntimeError("Bad position for Thing",
                                   position.x, position.y, position.z)

            self._position = mcpi.vec3.Vec3(position.x, position.y, position.z)

        if scene is None:
            # If no Scenes exists yet, create a new one
            if not World.scenes:
                Scene()  # Scene add itself to the World

            """ Use the default  Scene """
            self._scene = World.first_scene()

        # Add then thing built to the scene
        if parent is None:
            self._scene.add(self)

        # McThing version which created this Thing
        self._version = __version__

    @property
    def end_position(self):
        """ end position of the thing """
        return self._end_position

    @property
    def position(self):
        """ initial position of the thing """
        return self._position

    @property
    def parent(self):
        """ parent Thing in which this one is included """
        return self._position

    @property
    def scene(self):
        """ scene which this thing is included """
        return self._scene

    def add_child(self, child):
        """ Add a children to this Thing  """
        self._children.append(child)

    def set_block(self, pos, block_id, block_data=None):
        self._blocks_memory.set_block(pos, block_id, block_data)

    def set_blocks(self, init_pos, end_pos, block_id):
        """ Add a cuboid with the same block for all blocks and without specific data"""
        self._blocks_memory.set_blocks(init_pos, end_pos, block_id)

    def create(self):
        """
        Create the Thing in memory (BlocksMemory)
        :return:
        """

    def render(self):
        """
        Render the Thing from memory (BlocksMemory) to show it

        :return:
        """

        World.renderer.render(self._blocks_memory)
        for child in self._children:
            child.render()

    def build(self):
        """
        Build the thing and show it using the renderer at position coordinates

        :return:
        """

        self.create()
        self.render()

    def unbuild(self):
        """
        Unbuild the thing in Minecraft

        :return:
        """

        self._blocks_memory.fill(self._block_empty)
        self.render()
        self._blocks_memory.blocks = []

    def move(self, position):
        """
        Move the thing to a new position

        :param position: new position
        :return:
        """

        self.unbuild()
        self._position = position
        self.build()

    def rotate(self, degrees):
        """
        Rotate the thing in the x,z space using the blocks memory.

        :param degrees: degrees to rotate (90, 180, 270)
        :return:
        """

        self._blocks_memory.rotate(degrees, self.position)

        # Update the position and end_position after the rotation
        init_pos, end_pos = self._blocks_memory.find_init_end_pos()
        self._position = init_pos
        self._end_position = end_pos

    def flip_x(self):
        """
        Flip x-axis the things using the blocks memory.

        :return:
        """

        self._blocks_memory.flip_x(self.position)

        # Update the position and end_position after the rotation
        init_pos, end_pos = self._blocks_memory.find_init_end_pos()
        self._position = init_pos
        self._end_position = end_pos

    def to_schematic(self, file_path, blocks_data=False):
        """
        Convert the Thing to a Schematic Object

        :file_path: file in which to export the Thing in Schematic format
        :blocks_data: include blocks data (much slower)
        :return: the Schematic object
        """

        build_schematic_nbt(self.position, self.end_position, blocks_data).write_file(file_path)

    def add_decorator(self, decorator):
        """
        Add a new Decorator to be called once the Thing is decorated

        :param decorator: a Decorator to be called
        :return:
        """
        self._decorators.append(decorator)

    def decorate(self):
        """
        Call all decorators for the current Thing

        :return:
        """
        for decorator in self._decorators:
            decorator(self).decorate()
            for child in self._children:
                decorator(child).decorate()

    def find_bounding_box(self):
        """ Compute the bounding box of the Thing """

        return self.position, self.end_position



# Licensed under the terms of http://www.apache.org/licenses/LICENSE-2.0
# Author (©): Alvaro del Castillo

import mcpi

from .thing import Thing
from .house import House


class Town(Thing):

    houses = 4
    house_width = 5
    house_length = 5
    house_height = 3
    house_mirror = False
    space = 3
    """space between the town houses"""

    def create(self):

        init_x = self.position.x
        init_y = self.position.y
        init_z = self.position.z

        house_pos = mcpi.vec3.Vec3(init_x, init_y, init_z)

        for i in range(0, self.houses):
            house = House(house_pos, self)
            self.add_child(house)
            house.width = self.house_width
            house.length = self.house_length
            house.height = self.house_height
            house.block = self.block
            house.mirror = self.house_mirror
            house.create()
            house_pos.z += self.house_width + self.space

        # Fill the end_position
        end_x = init_x + self.house_length
        end_y = init_y + self.house_height
        end_z = house_pos.z - self.space

        self._end_position = mcpi.vec3.Vec3(end_x, end_y, end_z)



# Licensed under the terms of http://www.apache.org/licenses/LICENSE-2.0
# Author (©): Alvaro del Castillo
import logging
import sys

import mcpi
from mcpi.minecraft import Minecraft
from minecraftstuff import MinecraftDrawing

from .renderer import Renderer
from mcthings.blocks_memory import BlocksMemory


class _Server:
    """
    A Server manages the connection with the Minecraft server.

    Every World must have a Server in which built the World.
    """

    def __init__(self, host="localhost", port="4711"):
        self._host = host
        self._port = port

        self._mc = Minecraft.create(address=host, port=port)
        self._drawing = MinecraftDrawing(self._mc)

    @property
    def drawing(self):
        """ Connection to MinecraftDrawing (only used in Things built with MinecraftDrawing)"""
        return self._drawing

    @property
    def mc(self):
        """ Connection to Minecraft """
        return self._mc


class RaspberryPi(Renderer):
    """
    Renderer implemented using the Raspberry Pi Python API
    https://www.stuffaboutcode.com/p/minecraft-api-reference.html

    """

    def __init__(self, host, port):
        try:
            self.server = _Server(host, port)
        except mcpi.connection.RequestError:
            logging.error("Can't connect to Minecraft/Minetest server %s:%s" % (host, port))
            sys.exit(1)

    def render_cuboid_memory(self, memory):
        """ Render a memory with all blocks equal in a filled cuboid """
        block = memory.blocks[0]

        init_pos, end_pos = memory.find_init_end_pos()

        self.server.mc.setBlocks(init_pos.x, init_pos.y, init_pos.z,
                                 end_pos.x, end_pos.y, end_pos.z,
                                 block.id)

    def render_memory(self, memory):
        """ Render memory """

        for block in memory.blocks:
            if block.data is not None:
                self.server.mc.setBlock(block.pos.x, block.pos.y, block.pos.z, block.id, block.data)
            else:
                self.server.mc.setBlock(block.pos.x, block.pos.y, block.pos.z, block.id)

    def render(self, blocks_memory):
        if blocks_memory.memory_equal() and blocks_memory.is_cuboid():
            self.render_cuboid_memory(blocks_memory)
        else:
            self.render_memory(blocks_memory)

    def post_to_chat(self, message):
        self.server.mc.postToChat(message)

    def get_block(self, pos):
        return self.server.mc.getBlock(pos.x, pos.y, pos.z)

    def get_block_with_data(self, pos):
        return self.server.mc.getBlockWithData(pos.x, pos.y, pos.z)

    def get_blocks(self, init_pos, end_pos):
        return self.server.mc.getBlocks(init_pos.x, init_pos.y, init_pos.z, end_pos.x, end_pos.y, end_pos.z)

    def get_pos(self, entity):
        return self.server.mc.entity.getTilePos(
            self.server.mc.getPlayerEntityId(entity))
